"""
    *****************************************
    The Query Abstraction Layer(QAL) package.
    *****************************************

    QAL is designed to be a general library of data, data mining and database management functions.
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.

"""
# Version and release information used by Sphinx for documentation and setuptools for package generation. 
__version__ = '0.4'
__release__ = '0.4.0'
__copyright__ = '2010-2014, Nicklas Boerjesson'