"""
    ************************************
    qal.dal - Database Abstraction Layer
    ************************************
    
    The goal with the DAL package is to hide the connection-related differences between the most common RDBMS database backends.
    Supported platforms are (qal.dal.types) : MySQL, PostgreSQL, Oracle, DB2 and SQL server
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""