"""
    ********************************************************************
    qal.tools is a collection of tools for data management related tasks
    ********************************************************************
    These are used by the users of QAL, like the Optimal BPM data merge/migration tool.
    (See: https://sourceforge.net/projects/optimalbpm/files/Optimal%20BPM%20tools/)
    
    There are diff, merge and transform classes, but also nmap-based network 
    discovery of RDBMS hosts and a collection of functions for querying information from them.
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""