"""
    ****************************************
    qal.common - Commonly used functionality
    ****************************************

    This package gathers miscellaneous functionality that is used around QAL and the users of QAL. 
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""