"""
    ******************************
    qal.dataset - Data set classes
    ******************************
    The dataset package contains different data sets, all inheriting from the :py:class:qal.dataset.custom.custom class.
    They provide a generalized way of reading and writing to a multitude of data sources.
    Currenctly, there are: flat file, matrix, XPath(XML/HTML/XHTML) and spreadsheet(Excel-formats/.odt)
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""